REM
REM  Emulation core of the Zilog 80 CPU clone used in GameBoy consoles
REM  Part of the zBoy project
REM  Copyright (C) Mateusz Viste 2010
REM

TYPE CpuRegisters ' Note: IX and IY have been removed from the GB Z80 clone
  REM 8 bit registers
  A AS UBYTE   ' Accumulator
  B AS UBYTE   ' Commonly used as a 8bit counter
  C AS UBYTE
  D AS UBYTE
  E AS UBYTE
  F AS UBYTE   ' Flags
  H AS UBYTE
  L AS UBYTE
  'I AS UBYTE   ' Interrupt vector register
  'R AS UBYTE   ' Refresh register / might be used to generate random numbers
  REM 16 bit register - Note: to write/read composite register, use dedicated functions! (eg. ReadRegAF)
  PC AS UINTEGER ' Program counter (holds memory adress currently executed code). No function can change PC
  SP AS INTEGER ' The stack pointer. Holds current address of the top of the stack
END TYPE

DIM SHARED Register AS CpuRegisters
DIM SHARED InterruptsState AS UBYTE = 0  '  1 = ON  0 = OFF   (this is the IME register)
DIM SHARED HaltBug AS BYTE = 0  ' used to emulate the "HALT bug" (using HLT when IME is 0) -> The Smurfs won't work without that!
DIM SHARED HaltState AS BYTE = 0      ' 0 = CPU is running   1 = CPU is HALTed (wait until next interrupt)

#INCLUDE ONCE "cpu-timer.bi"      ' Include timer emulation
#INCLUDE ONCE "cpu-divider.bi"    ' Include divider emulation
#INCLUDE ONCE "cpu-microcode.bi"  ' add microinstruction set

#IFDEF DEBUGMODE
  DIM SHARED DebugCpuPause AS BYTE = 0   ' If set, pause at each CPU step
#ENDIF


#MACRO PopPCfromStack()
  STATIC AS UBYTE tUbyteBuff1, tUbyteBuff2
  PopFromStack(tUbyteBuff1, tUbyteBuff2)
  Register.PC = (tUbyteBuff1 SHL 8) OR tUbyteBuff2
#ENDMACRO

#MACRO PushtoStack16(x)  ' Push to stack a 16bit value
  PushToStack(((x) SHR 8), ((x) AND &b11111111))
#ENDMACRO


DIM SHARED IrqTempState AS UBYTE  ' Don't really need to be shared... but don't harm either :)
#MACRO CheckInterrupts()
  ' IME: Interrupt Master Enable (InterruptsState)
  ' IE: Interrupt Enable (&hFFFF)
  ' IF: Interrupt flag   (&hFF0F)
  IrqTempState = (MemoryInternalHiRAM(&hFFFF) AND IoRegisters(&hFF0F))    ' IE AND IF
  IF (IrqTempState > 0) AND (InterruptsState = 1) THEN   ' Check also InterruptsState (IME)
    IF (HaltState = 1) THEN
      HaltState = 0   ' If the CPU was HALTed, resume it
      Register.PC += 1
    END IF
    PushToStack16(Register.PC)            ' Save current address
    InterruptsState = 0   '  Reset the IME flag   * this is MANDATORY! *
    IF (IrqTempState AND &b00000001) = &b00000001 THEN  ' V-Blank
      PrintDebug("INT: VBLANK")
      Register.PC = &h40
      IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) AND &b11111110)  ' Reset the trigerred IF flag
    ELSEIF (IrqTempState AND &b00000010) = &b00000010 THEN  ' LCDC
      PrintDebug("INT: LCDC")
      Register.PC = &h48
      IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) AND &b11111101)  ' Reset the trigerred IF flag
    ELSEIF (IrqTempState AND &b00000100) = &b00000100 THEN  ' TIMER
      PrintDebug("INT: TIMER")
      Register.PC = &h50
      IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) AND &b11111011)  ' Reset the trigerred IF flag
    ELSEIF (IrqTempState AND &b00001000) = &b00001000 THEN  ' SERIAL
      PrintDebug("INT: SERIAL")
      Register.PC = &h58
      IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) AND &b11110111)  ' Reset the trigerred IF flag
    ELSEIF (IrqTempState AND &b00010000) = &b00010000 THEN ' Joypad
      PrintDebug("INT: JOYPAD")
      Register.PC = &h60
      IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) AND &b11101111)  ' Reset the trigerred IF flag
    END IF
  END IF
#ENDMACRO


SUB CpuExec()
  STATIC AS UBYTE UbyteBuff1, UbyteBuff2   ' Temporary buffer for various UBYTE values
  STATIC AS BYTE ByteBuff     ' Temporary buffer for various BYTE values
  STATIC AS UINTEGER UintBuff ' Temporary buffer for various UINTEGER values
  STATIC AS UBYTE UsedCycles
  STATIC AS UBYTE CpuInstruction ' holds current CPU instruction (bytecode)

  'IF TotalCycles > 1610000 THEN DebugCpuPause = 1   ' For debug purpose only! Enables stepbystep after a fixed time.
  'IF Register.PC = &h38 THEN DebugCpuPause = 1      ' For debug purpose only! Enables stepbystep when reach a specific address

  IF (Register.PC > &hFFFF) THEN
    PrintMsg("FATAL ERROR: PROGRAM COUNTER (PC) EXCEEDED FFFF!" + CHR(10,10) + "EXECUTED " & TotalCycles & " CYCLES")
    #IFDEF DEBUGMODE
      DebugOnScreen()
      DebugShowCpuRegisters()
    #ENDIF
    PressAnyKey()
    END(4)
  END IF

  MemoryRead(Register.PC, CpuInstruction)   ' Fetch the next opcode

  IF HaltBug = 1 THEN   ' Emulates the "HALT bug", as described in the Gameboy CPU manual at page 20
    HaltBug = 0         ' Some games are relying on this HALT bug, eg. "The Smurfs"
    Register.PC -= 1
  END IF

  'UsedCycles = 0          ' Should be removed later - no need to set it 0

  SELECT CASE AS CONST CpuInstruction
    CASE &h00   ' NOP (No operation)
      PrintDebug("NOP")  ' DEBUG
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h01   ' LD BC,nn  (load nn into BC)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD BC," & HEX(DwordVal(UbyteBuff1, UbyteBuff2)))  ' DEBUG
      WriteRegBC(UbyteBuff2, UbyteBuff1)
      Register.PC += 3
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h02   ' LD (BC),A   ' Load value of A into address at BC
      PrintDebug("LD (BC),A")  ' DEBUG
      MemoryWrite(ReadRegBC(), Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h03   ' INC BC (register BC+=1)
      PrintDebug("INC BC")  ' DEBUG
      IncReg16(Register.B, Register.C)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h04   ' INC B (register B+=1)
      PrintDebug("INC B")  ' DEBUG
      IncReg8(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h05   ' DEC B (register B-=1)
      PrintDebug("DEC B")  ' DEBUG
      DecReg8(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h06   ' LD B,n (load n into B)
      PrintDebug("LD B,n")  ' DEBUG
      MemoryRead(Register.PC + 1, Register.B)
      Register.PC += 2
      UsedCycles = 8
    CASE &h07   ' RLCA (Rotate reg A left, old bit 7 written to Carry flag)
      PrintDebug("RLCA")  '  DEBUG
      RLCA()
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h08   ' LD (nn),SP  (puts SP value at nn)
      'MemoryWrite16(DwordVal(MemoryRead(Register.PC + 1),MemoryRead(Register.PC + 2)), Register.SP)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD " & HEX(DwordVal(UbyteBuff1,UbyteBuff2)) & ",SP")
      MemoryWrite(DwordVal(UbyteBuff1,UbyteBuff2), (Register.SP AND &b11111111))
      MemoryWrite(DwordVal(UbyteBuff1,UbyteBuff2) + 1, (Register.SP SHR 8) AND &b11111111)
      Register.PC += 3
      UsedCycles = 20  ' that many CPU cycles should be spent on this instruction
    CASE &h09   ' ADD HL,BC
      PrintDebug("ADD HL,BC")
      AddToHL(ReadRegBC())
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h0A   ' LD A,(BC)  (put value from address BC to A)
      PrintDebug("LD A,(BC)")  ' DEBUG
      MemoryRead(ReadRegBC(), Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h0B   ' DEC BC (register BC-=1)
      PrintDebug("DEC BC")  ' DEBUG
      DecReg16(Register.B, Register.C)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h0C   ' INC C (register C+=1)
      PrintDebug("INC C")  ' DEBUG
      IncReg8(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h0D   ' DEC C (register C-=1)
      PrintDebug("DEC C")  ' DEBUG
      DecReg8(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h0E   ' LD C,n (load n into C)
      PrintDebug("LD C,n")  ' DEBUG
      MemoryRead(Register.PC + 1, Register.C)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h0F   ' RRCA (Rotate reg A right, old bit 0 written to Carry flag)
      PrintDebug("RRCA")  '  DEBUG
      RRCA()
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h10   ' STOP (opcode 10 00)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      IF UbyteBuff1 = &h00 THEN
          #IFNDEF DEBUGMODE
            PrintMsg("GOT A STOP INSTRUCTION AT " & HEX(Register.PC, 4))
            PressAnyKey()
          #ENDIF
          #IFDEF DEBUGMODE
            PrintDebug("Got a STOP at " & HEX(Register.PC, 4) & " (Opcode: 10 00)")
          #ENDIF
          QuitEmulator = 1
        ELSE
          #IFNDEF DEBUGMODE
            MemoryRead(Register.PC + 1, UbyteBuff1)
            PrintMsg("GOT AN UNKNOWN STOP-LIKE INSTRUCTION!" + CHR(10) + "ADDRESS: " + HEX(Register.PC, 4) + CHR(10) + "OPCODE: 10 " & HEX(UbyteBuff1, 2))
            PressAnyKey()
          #ENDIF
          #IFDEF DEBUGMODE
            PrintDebug("Got an unknown stop-like instruction at " & HEX(Register.PC, 4) & "!")
          #ENDIF
          QuitEmulator = 1
      END IF
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h11   ' LD DE,nn  (load nn into DE)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD DE," & HEX(DwordVal(UbyteBuff1, UbyteBuff2)))  ' DEBUG
      WriteRegDE(UbyteBuff2, UbyteBuff1)
      Register.PC += 3
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h12   ' LD (DE),A   ' Load value of A into address at DE
      PrintDebug("LD (DE),A")  ' DEBUG
      MemoryWrite(ReadRegDE(), Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h13   ' INC DE (register DE+=1)
      PrintDebug("INC DE")  ' DEBUG
      IncReg16(Register.D, Register.E)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h14   ' INC D (register D+=1)
      PrintDebug("INC D")  ' DEBUG
      IncReg8(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h15   ' DEC D (register D-=1)
      PrintDebug("DEC D")  ' DEBUG
      DecReg8(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h16   ' LD D,n (load n into D)
      PrintDebug("LD D,n")  ' DEBUG
      MemoryRead(Register.PC + 1, Register.D)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h17  ' RLA
      PrintDebug("RLA")  ' DEBUG
      RotateLeftCarry(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h18   ' JR n (jump to PC+n)
      Register.PC += 2  ' First increment the PC, and then jump
      MemoryRead(Register.PC - 1, UbyteBuff1)
      PrintDebug("JR " & UbyteToByte(UbyteBuff1))  ' DEBUG
      Register.PC += UbyteToByte(UbyteBuff1)
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h19   ' ADD HL,DE
      PrintDebug("ADD HL,DE")
      AddToHL(ReadRegDE())
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h1A   ' LD A,(DE)  (put value from address DE to A)
      PrintDebug("LD A,(DE)")
      MemoryRead(ReadRegDE(), Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h1B   ' DEC DE (register DE-=1)
      PrintDebug("DEC DE")  ' DEBUG
      DecReg16(Register.D, Register.E)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h1C   ' INC E (register E+=1)
      PrintDebug("INC E")  ' DEBUG
      IncReg8(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h1D   ' DEC E (register E-=1)
      PrintDebug("DEC E")  ' DEBUG
      DecReg8(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h1E   ' LD E,n (load n into E)
      PrintDebug("LD E,n")  ' DEBUG
      MemoryRead(Register.PC + 1, Register.E)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h1F  ' RRA
      PrintDebug("RRA")  ' DEBUG
      RotateRightCarry(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h20   ' JR NZ,n (jump to PC+n if Z flag is false)
      PrintDebug("JR NZ,n")  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagZ() = 0 THEN
        MemoryRead(Register.PC - 1, UbyteBuff1)
        Register.PC += UbyteToByte(UbyteBuff1)
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h21   ' LD HL,nn  (load nn into HL)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD HL," & HEX(DwordVal(UbyteBuff1, UbyteBuff2)))  ' DEBUG
      WriteRegHL(UbyteBuff2, UbyteBuff1)
      Register.PC += 3
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h22   ' LD (HL+),A (put A into memory addr HL, and increments HL)
      PrintDebug("LD (HL+),A")   ' DEBUG
      MemoryWrite(ReadRegHL(), Register.A)
      IncReg16(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h23   ' INC HL (register HL+=1)
      PrintDebug("INC HL")  ' DEBUG
      IncReg16(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h24   ' INC H (register H+=1)
      PrintDebug("INC H")  ' DEBUG
      IncReg8(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h25   ' DEC H (register H-=1)
      PrintDebug("DEC H")  ' DEBUG
      DecReg8(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h26   ' LD H,n (load n into H)
      PrintDebug("LD H,n")  ' DEBUG
      MemoryRead(Register.PC + 1, Register.H)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h27   ' DAA
      PrintDebug("DAA")  ' DEBUG
      AdjustDAA()
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h28   ' JR Z,n (jump to PC+n if Z flag is true)
      PrintDebug("JR Z,n")  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagZ() = 1 THEN
        MemoryRead(Register.PC - 1, UbyteBuff1)
        Register.PC += UbyteToByte(UbyteBuff1)
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h29   ' ADD HL,HL
      PrintDebug("ADD HL,HL")
      AddToHL(ReadRegHL())
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h2A   ' LD A,(HL+) (Put value @HL into A, and increment HL by 1)
      PrintDebug("LD A,(HL+)")    ' DEBUG
      MemoryRead(ReadRegHL(), Register.A)
      IncReg16(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h2B   ' DEC HL (register HL-=1)
      PrintDebug("DEC HL")  ' DEBUG
      DecReg16(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h2C   ' INC L (register L+=1)
      PrintDebug("INC L")  ' DEBUG
      IncReg8(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h2D   ' DEC L (register L-=1)
      PrintDebug("DEC L")  ' DEBUG
      DecReg8(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h2E   ' LD L,n (load n into L)
      PrintDebug("LD L,n")  ' DEBUG
      MemoryRead(Register.PC + 1, Register.L)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h2F   ' CPL (complements the A register, that is flips all bits)
      PrintDebug("CPL")
      SetFlagN() : SetFlagH()
      Register.A = Register.A XOR &b11111111  ' Flip all bits
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h30   ' JR NC,n (jump to PC+n if C flag is false)
      PrintDebug("JR NC,n")  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagC() = 0 THEN
        MemoryRead(Register.PC - 1, UbyteBuff1)
        Register.PC += UbyteToByte(UbyteBuff1)
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h31   ' LD SP,nn
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD SP," & HEX(DwordVal(UbyteBuff1, UbyteBuff2)))  ' DEBUG
      Register.SP = DwordVal(UbyteBuff1, UbyteBuff2)
      Register.PC += 3
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h32   ' LD (HL-),A   (put A into address at HL, then decrement HL)
      PrintDebug("LD (HL-),A")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.A)
      DecReg16(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h33   ' INC SP (register SP+=1)
      PrintDebug("INC SP")  ' DEBUG
      IF (Register.SP = 65535) THEN
          Register.SP = 0
        ELSE
          Register.SP += 1
      END IF
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h34   ' INC (HL) (value at (HL) is incremented)
      PrintDebug("INC (HL)")   ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      IncReg8(UbyteBuff1)
      MemoryWrite(ReadRegHL(), UbyteBuff1)
      Register.PC += 1
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h35   ' DEC (HL) -> (HL) -= 1
      PrintDebug("DEC (HL)")   ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)  ' Save old value at (HL)
      DecReg8(UbyteBuff1)
      MemoryWrite(ReadRegHL(), UbyteBuff1)
      Register.PC += 1
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h36   ' LD (HL),n   ' Load value of n into address at HL
      PrintDebug("LD (HL),n")  ' DEBUG
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryWrite(ReadRegHL(), UbyteBuff1)
      Register.PC += 2
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &h37   ' SCF (Set Carry Flag)
      PrintDebug("SCF")  ' DEBUG
      ResetFlagN() : ResetFlagH() : SetFlagC()
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h38   ' JR C,n (jump to PC+n if C flag is true)
      PrintDebug("JR C,n")  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagC() = 1 THEN
        MemoryRead(Register.PC - 1, UbyteBuff1)
        Register.PC += UbyteToByte(UbyteBuff1)
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h39   ' ADD HL,SP
      PrintDebug("ADD HL,SP")
      AddToHL(Register.SP)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h3A   ' LD A,(HL-)   (put value at address HL into A, and decrement HL)
      PrintDebug("LD A,(HL-)")
      MemoryRead(ReadRegHL(), Register.A)
      DecReg16(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h3B   ' DEC SP (register SP-=1)
      PrintDebug("DEC SP")  ' DEBUG
      IF (Register.SP = 0) THEN
          Register.SP = 65535
        ELSE
          Register.SP -= 1
      END IF
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h3C   ' INC A (register A+=1)
      PrintDebug("INC A")  ' DEBUG
      IncReg8(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h3D   ' DEC A (register A-=1)
      PrintDebug("DEC A")  ' DEBUG
      DecReg8(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h3E   ' LD A,n  (put n into A)
      MemoryRead(Register.PC + 1, Register.A)
      PrintDebug("LD A," & Register.A)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h3F   ' CCF  (Complement Carry Flag)
      PrintDebug("CCF")
      ResetFlagN() : ResetFlagH()
      IF GetFlagC() = 0 THEN SetFlagC() ELSE ResetFlagC()
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h40   ' LD B,B     ' Put value B into B
      PrintDebug("LD B,B")  ' DEBUG
      Register.B = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h41   ' LD B,C     ' Put value C into B
      PrintDebug("LD B,C")  ' DEBUG
      Register.B = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h42   ' LD B,D     ' Put value D into B
      PrintDebug("LD B,D")  ' DEBUG
      Register.B = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h43   ' LD B,E     ' Put value E into B
      PrintDebug("LD B,E")  ' DEBUG
      Register.B = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h44   ' LD B,H     ' Put value H into B
      PrintDebug("LD B,H")  ' DEBUG
      Register.B = Register.H
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h45   ' LD B,L     ' Put value L into B
      PrintDebug("LD B,L")  ' DEBUG
      Register.B = Register.L
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h46   ' LD B,(HL)  (put value from address HL to B)
      PrintDebug("LD B,(HL)")
      MemoryRead(ReadRegHL(), Register.B)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h47   ' LD B,A     ' Put value A into B
      PrintDebug("LD B,A")  ' DEBUG
      Register.B = Register.A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h48   ' LD C,B     ' Put value B into C
      PrintDebug("LD C,B")  ' DEBUG
      Register.C = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h49   ' LD C,C     ' Put value C into C
      PrintDebug("LD C,C")  ' DEBUG
      Register.C = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h4A   ' LD C,D     ' Put value D into C
      PrintDebug("LD C,D")  ' DEBUG
      Register.C = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h4B   ' LD C,E     ' Put value E into C
      PrintDebug("LD C,E")  ' DEBUG
      Register.C = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h4C   ' LD C,H     ' Put value H into C
      PrintDebug("LD C,H")  ' DEBUG
      Register.C = Register.H
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h4D   ' LD C,L     ' Put value L into C
      PrintDebug("LD C,L")  ' DEBUG
      Register.C = Register.L
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h4E   ' LD C,(HL)  (put value from address HL to C)
      PrintDebug("LD C,(HL)")
      MemoryRead(ReadRegHL(), Register.C)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h4F   ' LD C,A     ' Put value A into C
      PrintDebug("LD C,A")  ' DEBUG
      Register.C = Register.A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h50   ' LD D,B     ' Put value B into D
      PrintDebug("LD D,B")  ' DEBUG
      Register.D = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h51   ' LD D,C     ' Put value C into D
      PrintDebug("LD D,C")  ' DEBUG
      Register.D = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h52   ' LD D,D     ' Put value D into D
      PrintDebug("LD D,D")  ' DEBUG
      Register.D = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h53   ' LD D,E     ' Put value E into D
      PrintDebug("LD D,E")  ' DEBUG
      Register.D = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h54   ' LD D,H     ' Put value H into D
      PrintDebug("LD D,H")  ' DEBUG
      Register.D = Register.H
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h55   ' LD D,L     ' Put value L into D
      PrintDebug("LD D,L")  ' DEBUG
      Register.D = Register.L
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h56   ' LD D,(HL)  (put value from address HL to D)
      PrintDebug("LD D,(HL)")
      MemoryRead(ReadRegHL(), Register.D)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h57   ' LD D,A     ' Put value A into D
      PrintDebug("LD D,A")  ' DEBUG
      Register.D = Register.A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h58   ' LD E,B     ' Put value B into E
      PrintDebug("LD E,B")  ' DEBUG
      Register.E = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h59   ' LD E,C     ' Put value C into E
      PrintDebug("LD E,C")  ' DEBUG
      Register.E = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h5A   ' LD E,D     ' Put value D into E
      PrintDebug("LD E,D")  ' DEBUG
      Register.E = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h5B   ' LD E,E     ' Put value E into E
      PrintDebug("LD E,E")  ' DEBUG
      Register.E = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h5C   ' LD E,H     ' Put value H into E
      PrintDebug("LD E,H")  ' DEBUG
      Register.E = Register.H
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h5D   ' LD E,L     ' Put value L into E
      PrintDebug("LD E,L")  ' DEBUG
      Register.E = Register.L
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h5E   ' LD E,(HL)  (put value from address HL to E)
      PrintDebug("LD E,(HL)")
      MemoryRead(ReadRegHL(), Register.E)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h5F   ' LD E,A     ' Put value A into E
      PrintDebug("LD E,A")  ' DEBUG
      Register.E = Register.A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h60   ' LD H,B     ' Put value B into H
      PrintDebug("LD H,B")  ' DEBUG
      Register.H = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h61   ' LD H,C     ' Put value C into H
      PrintDebug("LD H,C")  ' DEBUG
      Register.H = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h62   ' LD H,D     ' Put value D into H
      PrintDebug("LD H,D")  ' DEBUG
      Register.H = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h63   ' LD H,E     ' Put value E into H
      PrintDebug("LD H,E")  ' DEBUG
      Register.H = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h64   ' LD H,H     ' Put value H into H
      PrintDebug("LD H,H")  ' DEBUG
      'Register.H = Register.H   ' Commented out, because there is no point in doing that
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h65   ' LD H,L     ' Put value L into H
      PrintDebug("LD H,L")  ' DEBUG
      Register.H = Register.L
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h66   ' LD H,(HL)  (put value from address HL to H)
      PrintDebug("LD H,(HL)")
      MemoryRead(ReadRegHL(), UbyteBuff1)
      Register.H = UbyteBuff1   ' Don't try to work on H directly! It would change also HL!
      'Register.H = MemoryRead(ReadRegHL())
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h67   ' LD H,A     ' Put value A into H
      PrintDebug("LD H,A")  ' DEBUG
      Register.H = Register.A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h68   ' LD L,B     ' Put value B into L
      PrintDebug("LD L,B")  ' DEBUG
      Register.L = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h69   ' LD L,C     ' Put value C into L
      PrintDebug("LD L,C")  ' DEBUG
      Register.L = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h6A   ' LD L,D     ' Put value D into L
      PrintDebug("LD L,D")  ' DEBUG
      Register.L = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h6B   ' LD L,E     ' Put value E into L
      PrintDebug("LD L,E")  ' DEBUG
      Register.L = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h6C   ' LD L,H     ' Put value H into L
      PrintDebug("LD L,H")  ' DEBUG
      Register.L = Register.H
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h6D   ' LD L,L     ' Put value L into L
      PrintDebug("LD L,L")  ' DEBUG
      'Register.L = Register.L   ' Commented out, because there is no point in doing such operation
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h6E   ' LD L,(HL)  (put value from address HL to L)
      PrintDebug("LD L,(HL)")
      MemoryRead(ReadRegHL(), UbyteBuff1)
      Register.L = UbyteBuff1    ' Don't try to work directly on L! It would also change HL!
      'Register.L = MemoryRead(ReadRegHL())
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h6F   ' LD L,A     ' Put value A into L
      PrintDebug("LD L,A")  ' DEBUG
      Register.L = Register.A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h70   ' LD (HL),B   ' Load value of B into address at HL
      PrintDebug("LD (HL),B")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.B)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h71   ' LD (HL),C   ' Load value of C into address at HL
      PrintDebug("LD (HL),C")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.C)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h72   ' LD (HL),D   ' Load value of D into address at HL
      PrintDebug("LD (HL),D")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.D)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h73   ' LD (HL),E   ' Load value of E into address at HL
      PrintDebug("LD (HL),E")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.E)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h74   ' LD (HL),H   ' Load value of H into address at HL
      PrintDebug("LD (HL),H")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.H)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h75   ' LD (HL),L   ' Load value of L into address at HL
      PrintDebug("LD (HL),L")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.L)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h76   ' HALT   ' HALTs the CPU (only if interrupts are permitted) until next interrupt
      REM  Note: I don't increment PC here, to make the game loop on the HALT op.
      REM        PC will be incremented if an interrupt occurs.
      PrintDebug("Halt")
      IF (InterruptsState = 1) THEN
          HaltState = 1
        ELSE
          Register.PC += 1
          HaltBug = 1
      END IF
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h77   ' LD (HL),A   ' Load value of A into address at HL
      PrintDebug("LD (HL),A")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h78   ' LD A,B  (put B into A)
      PrintDebug("LD A,B")   ' DEBUG
      Register.A = Register.B
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h79   ' LD A,C  (put C into A)
      PrintDebug("LD A,C")   ' DEBUG
      Register.A = Register.C
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h7A   ' LD A,D  (put D into A)
      PrintDebug("LD A,D")   ' DEBUG
      Register.A = Register.D
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h7B   ' LD A,E  (put E into A)
      PrintDebug("LD A,E")   ' DEBUG
      Register.A = Register.E
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h7C   ' LD A,H  (put H into A)
      PrintDebug("LD A,H")   ' DEBUG
      Register.A = Register.H
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h7D   ' LD A,L  (put L into A)
      PrintDebug("LD A,L")   ' DEBUG
      Register.A = Register.L
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h7E   ' LD A,(HL)  (put value from address HL to A)
      PrintDebug("LD A,(HL)")
      MemoryRead(ReadRegHL(), Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h7F   ' LD A,A  (put A into A)
      PrintDebug("LD A,A")   ' DEBUG
      'Register.A = Register.A   ' Commented out, because there is no point in doing such operation
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h80  ' ADD A,B  (A+=B)
      PrintDebug("ADD A,B")
      AddToA(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h81  ' ADD A,C  (A+=C)
      PrintDebug("ADD A,C")
      AddToA(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h82  ' ADD A,D  (A+=D)
      PrintDebug("ADD A,D")
      AddToA(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h83  ' ADD A,E  (A+=E)
      PrintDebug("ADD A,E")
      AddToA(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h84  ' ADD A,H  (A+=H)
      PrintDebug("ADD A,H")
      AddToA(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h85  ' ADD A,L  (A+=L)
      PrintDebug("ADD A,L")
      AddToA(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h86  ' ADD A,(HL)  (A+=[HL])
      PrintDebug("ADD A,(HL)")
      MemoryRead(ReadRegHL(), UbyteBuff1)
      AddToA(UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h87  ' ADD A,A  (A+=A)
      PrintDebug("ADD A,A")
      AddToA(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h88   ' ADC A,B
      PrintDebug("ADC A,B")   ' DEBUG
      AdcA(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h89   ' ADC A,C
      PrintDebug("ADC A,C")   ' DEBUG
      AdcA(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h8A   ' ADC A,D
      PrintDebug("ADC A,D")   ' DEBUG
      AdcA(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h8B   ' ADC A,E
      PrintDebug("ADC A,E")   ' DEBUG
      AdcA(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h8C   ' ADC A,H
      PrintDebug("ADC A,H")   ' DEBUG
      AdcA(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h8D   ' ADC A,L
      PrintDebug("ADC A,L")   ' DEBUG
      AdcA(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h8E   ' ADC A,(HL)
      PrintDebug("ADC A,(HL)")   ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      AdcA(UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h8F   ' ADC A,A
      PrintDebug("ADC A,A")   ' DEBUG
      AdcA(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h90  ' SUB B
      PrintDebug("SUB B")  ' DEBUG
      SubValFromReg8(Register.A, Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h91  ' SUB C
      PrintDebug("SUB C")  ' DEBUG
      SubValFromReg8(Register.A, Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h92  ' SUB D
      PrintDebug("SUB D")  ' DEBUG
      SubValFromReg8(Register.A, Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h93  ' SUB E
      PrintDebug("SUB E")  ' DEBUG
      SubValFromReg8(Register.A, Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h94  ' SUB H
      PrintDebug("SUB H")  ' DEBUG
      SubValFromReg8(Register.A, Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h95  ' SUB L
      PrintDebug("SUB L")  ' DEBUG
      SubValFromReg8(Register.A, Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h96  ' SUB (HL)  (sub content at address HL from A)
      PrintDebug("SUB (HL)")  ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      SubValFromReg8(Register.A, UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h97  ' SUB A
      PrintDebug("SUB A")  ' DEBUG
      SubValFromReg8(Register.A, Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h98  ' SBC A,B
      PrintDebug("SBC A,B")  ' DEBUG
      SbcA(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h99  ' SBC A,C
      PrintDebug("SBC A,C")  ' DEBUG
      SbcA(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h9A  ' SBC A,D
      PrintDebug("SBC A,D")  ' DEBUG
      SbcA(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h9B  ' SBC A,E
      PrintDebug("SBC A,E")  ' DEBUG
      SbcA(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h9C  ' SBC A,H
      PrintDebug("SBC A,H")  ' DEBUG
      SbcA(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h9D  ' SBC A,L
      PrintDebug("SBC A,L")  ' DEBUG
      SbcA(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &h9E  ' SBC A,(HL)
      PrintDebug("SBC A,(HL)")  ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      SbcA(UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &h9F  ' SBC A,A
      PrintDebug("SBC A,A")  ' DEBUG
      SbcA(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA0   ' AND B (A = A AND B)
      PrintDebug("AND B")  ' DEBUG
      AndA(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA1   ' AND C (A = A AND C)
      PrintDebug("AND C")  ' DEBUG
      AndA(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA2   ' AND D (A = A AND D)
      PrintDebug("AND D")  ' DEBUG
      AndA(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA3   ' AND E (A = A AND E)
      PrintDebug("AND E")  ' DEBUG
      AndA(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA4   ' AND H (A = A AND H)
      PrintDebug("AND H")  ' DEBUG
      AndA(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA5   ' AND L (A = A AND L)
      PrintDebug("AND L")  ' DEBUG
      AndA(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA6   ' AND (HL) (A = A AND [HL])
      PrintDebug("AND (HL)")  ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      AndA(UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hA7   ' AND A (A = A AND A)
      PrintDebug("AND A")  ' DEBUG
      AndA(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA8   ' XOR B (A = A XOR B)
      PrintDebug("XOR B")  ' DEBUG
      XorA(Register.B) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hA9   ' XOR C (A = A XOR C)
      PrintDebug("XOR C")  ' DEBUG
      XorA(Register.C) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hAA   ' XOR D (A = A XOR D)
      PrintDebug("XOR D")  ' DEBUG
      XorA(Register.D) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hAB   ' XOR E (A = A XOR E)
      PrintDebug("XOR E")  ' DEBUG
      XorA(Register.E) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hAC   ' XOR H (A = A XOR H)
      PrintDebug("XOR H")  ' DEBUG
      XorA(Register.H) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hAD   ' XOR L (A = A XOR L)
      PrintDebug("XOR L")  ' DEBUG
      XorA(Register.L) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hAE   ' XOR (HL)  (A = A XOR [HL])
      PrintDebug("XOR (HL)") ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      XorA(UbyteBuff1) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hAF   ' XOR A (A = A XOR A)
      PrintDebug("XOR A")  ' DEBUG
      XorA(Register.A) ' Xor with A, result in A
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB0   ' OR B  (A = A OR B)
      PrintDebug("OR B")   ' DEBUG
      OrA(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB1   ' OR C  (A = A OR C)
      PrintDebug("OR C")   ' DEBUG
      OrA(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB2   ' OR D  (A = A OR D)
      PrintDebug("OR D")   ' DEBUG
      OrA(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB3   ' OR E  (A = A OR E)
      PrintDebug("OR E")   ' DEBUG
      OrA(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB4   ' OR H  (A = A OR H)
      PrintDebug("OR H")   ' DEBUG
      OrA(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB5   ' OR L  (A = A OR L)
      PrintDebug("OR L")   ' DEBUG
      OrA(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB6   ' OR (HL)  (A = A OR (HL))
      PrintDebug("OR (HL)")   ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      OrA(UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hB7   ' OR A  (A = A OR A)
      PrintDebug("OR A")   ' DEBUG
      OrA(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB8   ' CP B (compare A with B)
      PrintDebug("CP B") ' DEBUG
      CmpA(Register.B)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hB9   ' CP C (compare A with C)
      PrintDebug("CP C") ' DEBUG
      CmpA(Register.C)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hBA   ' CP D (compare A with D)
      PrintDebug("CP D") ' DEBUG
      CmpA(Register.D)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hBB   ' CP E (compare A with E)
      PrintDebug("CP E") ' DEBUG
      CmpA(Register.E)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hBC   ' CP H (compare A with H)
      PrintDebug("CP H") ' DEBUG
      CmpA(Register.H)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hBD   ' CP L (compare A with L)
      PrintDebug("CP L") ' DEBUG
      CmpA(Register.L)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hBE   ' CP (HL) (compare A with value at address [HL])
      PrintDebug("CP (HL)") ' DEBUG
      MemoryRead(ReadRegHL(), UbyteBuff1)
      CmpA(UbyteBuff1)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hBF   ' CP A (compare A with A)
      PrintDebug("CP A") ' DEBUG
      CmpA(Register.A)
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hC0   ' RET NZ   ' RET if Z = 0
      PrintDebug("RET NZ")
      IF GetFlagZ() = 0 THEN
          PopPCfromStack()
        ELSE
          Register.PC += 1
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hC1   ' POP BC (Write the value in stack into BC)
      PrintDebug("POP BC")  ' DEBUG
      PopFromStack(Register.B, Register.C)
      Register.PC += 1
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hC2   ' JP NZ,nn (jump if Z=0)
      PrintDebug("JP NZ,nn") ' DEBUG
      IF GetFlagZ() = 0 THEN  ' Jump
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hC3   ' JP nn (Unconditional jump)
      PrintDebug("JP nn") ' DEBUG
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hC4   ' CALL NZ,nn (call if Z=0)
      PrintDebug("CALL NZ,nn") ' DEBUG
      IF GetFlagZ() = 0 THEN  ' Call
          PushToStack16(Register.PC + 3)           ' First save the current value of the PC (+3 bytes)
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
        ELSE   ' Don't call
          Register.PC += 3
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hC5   ' PUSH BC
      PrintDebug("PUSH BC")  ' DEBUG
      PushToStack(Register.B, Register.C)
      Register.PC += 1
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hC6  ' ADD A,n  (A+=n)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("ADD A," & HEX(UbyteBuff1, 2))
      AddToA(UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hC7   ' RST 00h
      PrintDebug("RST 00h")   ' DEBUG
      PushToStack16(Register.PC + 1)            ' Save current address (+1)
      Register.PC = &h00  ' Jump to 00h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hC8   ' RET Z   ' RET if Z = 1
      PrintDebug("RET Z")
      IF GetFlagZ() = 0 THEN
          Register.PC += 1
        ELSE
          PopPCfromStack()
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hC9   ' RET
      PrintDebug("RET")
      PopPCfromStack()
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hCA   ' JP Z,nn (jump if Z=1)
      PrintDebug("JP Z,nn") ' DEBUG
      IF GetFlagZ() = 0 THEN  ' Don't jump
          Register.PC += 3
        ELSE   ' Jump
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hCB  ' Here we have a CBxx opcode...
      MemoryRead(Register.PC + 1, CpuInstruction)  '  Let's check the second byte!
      SELECT CASE AS CONST CpuInstruction
        CASE &h00  ' RLC B
          PrintDebug("RLC B")  ' DEBUG
          RotateRLC(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h01  ' RLC C
          PrintDebug("RLC C")  ' DEBUG
          RotateRLC(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h02  ' RLC D
          PrintDebug("RLC D")  ' DEBUG
          RotateRLC(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h03  ' RLC E
          PrintDebug("RLC E")  ' DEBUG
          RotateRLC(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h04  ' RLC H
          PrintDebug("RLC H")  ' DEBUG
          RotateRLC(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h05  ' RLC L
          PrintDebug("RLC L")  ' DEBUG
          RotateRLC(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h06  ' RLC (HL)
          PrintDebug("RLC (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          RotateRLC(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h07  ' RLC A
          PrintDebug("RLC A")  ' DEBUG
          RotateRLC(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h08  ' RRC B
          PrintDebug("RRC B")  ' DEBUG
          RotateRRC(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h09  ' RRC C
          PrintDebug("RRC C")  ' DEBUG
          RotateRRC(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h0A  ' RRC D
          PrintDebug("RRC D")  ' DEBUG
          RotateRRC(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h0B  ' RRC E
          PrintDebug("RRC E")  ' DEBUG
          RotateRRC(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h0C  ' RRC H
          PrintDebug("RRC H")  ' DEBUG
          RotateRRC(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h0D  ' RRC L
          PrintDebug("RRC L")  ' DEBUG
          RotateRRC(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h0E  ' RRC (HL)
          PrintDebug("RRC (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          RotateRRC(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h0F  ' RRC A
          PrintDebug("RRC A")  ' DEBUG
          RotateRRC(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h10  ' RL B
          PrintDebug("RL B")  ' DEBUG
          RotateLeftCarry(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h11  ' RL C
          PrintDebug("RL C")  ' DEBUG
          RotateLeftCarry(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h12  ' RL D
          PrintDebug("RL D")  ' DEBUG
          RotateLeftCarry(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h13  ' RL E
          PrintDebug("RL E")  ' DEBUG
          RotateLeftCarry(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h14  ' RL H
          PrintDebug("RL H")  ' DEBUG
          RotateLeftCarry(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h15  ' RL L
          PrintDebug("RL L")  ' DEBUG
          RotateLeftCarry(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h16  ' RL (HL)
          PrintDebug("RL (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          RotateLeftCarry(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h17  ' RL A
          PrintDebug("RL A")  ' DEBUG
          RotateLeftCarry(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h18  ' RR B
          PrintDebug("RR B")  ' DEBUG
          RotateRightCarry(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h19  ' RR C
          PrintDebug("RR C")  ' DEBUG
          RotateRightCarry(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h1A  ' RR D
          PrintDebug("RR D")  ' DEBUG
          RotateRightCarry(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h1B  ' RR E
          PrintDebug("RR E")  ' DEBUG
          RotateRightCarry(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h1C  ' RR H
          PrintDebug("RR H")  ' DEBUG
          RotateRightCarry(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h1D  ' RR L
          PrintDebug("RR L")  ' DEBUG
          RotateRightCarry(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h1E  ' RR (HL)
          PrintDebug("RR (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          RotateRightCarry(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h1F  ' RR A
          PrintDebug("RR A")  ' DEBUG
          RotateRightCarry(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h20  ' SLA B
          PrintDebug("SLA B")  ' DEBUG
          ShiftSLA(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h21  ' SLA C
          PrintDebug("SLA C")  ' DEBUG
          ShiftSLA(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h22  ' SLA D
          PrintDebug("SLA D")  ' DEBUG
          ShiftSLA(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h23  ' SLA E
          PrintDebug("SLA E")  ' DEBUG
          ShiftSLA(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h24  ' SLA H
          PrintDebug("SLA H")  ' DEBUG
          ShiftSLA(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h25  ' SLA L
          PrintDebug("SLA L")  ' DEBUG
          ShiftSLA(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h26  ' SLA (HL)
          PrintDebug("SLA (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ShiftSLA(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h27  ' SLA A
          PrintDebug("SLA A")  ' DEBUG
          ShiftSLA(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h28  ' SRA B
          PrintDebug("SRA B")  ' DEBUG
          ShiftSRA(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h29  ' SRA C
          PrintDebug("SRA C")  ' DEBUG
          ShiftSRA(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h2A  ' SRA D
          PrintDebug("SRA D")  ' DEBUG
          ShiftSRA(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h2B  ' SRA E
          PrintDebug("SRA E")  ' DEBUG
          ShiftSRA(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h2C  ' SRA H
          PrintDebug("SRA H")  ' DEBUG
          ShiftSRA(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h2D  ' SRA L
          PrintDebug("SRA L")  ' DEBUG
          ShiftSRA(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h2E  ' SRA (HL)
          PrintDebug("SRA (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ShiftSRA(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h2F  ' SRA A
          PrintDebug("SRA A")  ' DEBUG
          ShiftSRA(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h30  ' SWAP B (swaps high and low nibbles of B)
          PrintDebug("SWAP B")  ' DEBUG
          SwapUbyte(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h31  ' SWAP C (swaps high and low nibbles of C)
          PrintDebug("SWAP C")  ' DEBUG
          SwapUbyte(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h32  ' SWAP D (swaps high and low nibbles of D)
          PrintDebug("SWAP D")  ' DEBUG
          SwapUbyte(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h33  ' SWAP E (swaps high and low nibbles of E)
          PrintDebug("SWAP E")  ' DEBUG
          SwapUbyte(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h34  ' SWAP H (swaps high and low nibbles of H)
          PrintDebug("SWAP H")  ' DEBUG
          SwapUbyte(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h35  ' SWAP L (swaps high and low nibbles of L)
          PrintDebug("SWAP L")  ' DEBUG
          SwapUbyte(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h36  ' SWAP (HL) (swaps high and low nibbles of value at address HL)
          PrintDebug("SWAP (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SwapUbyte(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h37  ' SWAP A (swaps high and low nibbles of A)
          PrintDebug("SWAP A")  ' DEBUG
          SwapUbyte(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h38  ' SRL B  (Shift Right B)
          PrintDebug("SRL B")  ' DEBUG
          ShiftRightSRL(Register.B)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h39  ' SRL C  (Shift Right C)
          PrintDebug("SRL C")  ' DEBUG
          ShiftRightSRL(Register.C)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h3A  ' SRL D  (Shift Right D)
          PrintDebug("SRL D")  ' DEBUG
          ShiftRightSRL(Register.D)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h3B  ' SRL E  (Shift Right E)
          PrintDebug("SRL E")  ' DEBUG
          ShiftRightSRL(Register.E)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h3C  ' SRL H  (Shift Right H)
          PrintDebug("SRL H")  ' DEBUG
          ShiftRightSRL(Register.H)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h3D  ' SRL L  (Shift Right L)
          PrintDebug("SRL L")  ' DEBUG
          ShiftRightSRL(Register.L)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h3E  ' SRL (HL)  (Shift Right value at address HL)
          PrintDebug("SRL (HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ShiftRightSRL(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)   ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h3F  ' SRL A  (Shift Right A)
          PrintDebug("SRL A")  ' DEBUG
          ShiftRightSRL(Register.A)  ' Shift right and set some flags
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h40 ' BIT 0,B (Test bit 0 of register B)
          PrintDebug("BIT 0,B")  ' DEBUG
          TestBit0(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h41 ' BIT 0,C (Test bit 0 of register C)
          PrintDebug("BIT 0,C")  ' DEBUG
          TestBit0(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h42 ' BIT 0,D (Test bit 0 of register D)
          PrintDebug("BIT 0,D")  ' DEBUG
          TestBit0(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h43 ' BIT 0,E (Test bit 0 of register E)
          PrintDebug("BIT 0,E")  ' DEBUG
          TestBit0(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h44 ' BIT 0,H (Test bit 0 of register H)
          PrintDebug("BIT 0,H")  ' DEBUG
          TestBit0(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h45 ' BIT 0,L (Test bit 0 of register L)
          PrintDebug("BIT 0,L")  ' DEBUG
          TestBit0(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h46 ' BIT 0,(HL) (Test bit 0 of value at address HL)
          PrintDebug("BIT 0,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit0(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h47 ' BIT 0,A (Test bit 0 of register A)
          PrintDebug("BIT 0,A")  ' DEBUG
          TestBit0(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h48 ' BIT 1,B (Test bit 1 of register B)
          PrintDebug("BIT 1,B")  ' DEBUG
          TestBit1(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h49 ' BIT 1,C (Test bit 1 of register C)
          PrintDebug("BIT 1,C")  ' DEBUG
          TestBit1(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h4A ' BIT 1,D (Test bit 1 of register D)
          PrintDebug("BIT 1,D")  ' DEBUG
          TestBit1(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h4B ' BIT 1,E (Test bit 1 of register E)
          PrintDebug("BIT 1,E")  ' DEBUG
          TestBit1(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h4C ' BIT 1,H (Test bit 1 of register H)
          PrintDebug("BIT 1,H")  ' DEBUG
          TestBit1(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h4D ' BIT 1,L (Test bit 1 of register L)
          PrintDebug("BIT 1,L")  ' DEBUG
          TestBit1(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h4E ' BIT 1,(HL) (Test bit 1 of value at address HL)
          PrintDebug("BIT 1,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit1(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h4F ' BIT 1,A (Test bit 1 of register A)
          PrintDebug("BIT 1,A")  ' DEBUG
          TestBit1(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h50 ' BIT 2,B (Test bit 2 of register B)
          PrintDebug("BIT 2,B")  ' DEBUG
          TestBit2(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h51 ' BIT 2,C (Test bit 2 of register C)
          PrintDebug("BIT 2,C")  ' DEBUG
          TestBit2(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h52 ' BIT 2,D (Test bit 2 of register D)
          PrintDebug("BIT 2,D")  ' DEBUG
          TestBit2(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h53 ' BIT 2,E (Test bit 2 of register E)
          PrintDebug("BIT 2,E")  ' DEBUG
          TestBit2(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h54 ' BIT 2,H (Test bit 2 of register H)
          PrintDebug("BIT 2,H")  ' DEBUG
          TestBit2(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h55 ' BIT 2,L (Test bit 2 of register L)
          PrintDebug("BIT 2,L")  ' DEBUG
          TestBit2(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h56 ' BIT 2,(HL) (Test bit 2 of value at [HL])
          PrintDebug("BIT 2,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit2(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h57 ' BIT 2,A (Test bit 2 of register A)
          PrintDebug("BIT 2,A")  ' DEBUG
          TestBit2(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h58 ' BIT 3,B (Test bit 3 of register B)
          PrintDebug("BIT 3,B")  ' DEBUG
          TestBit3(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h59 ' BIT 3,C (Test bit 3 of register C)
          PrintDebug("BIT 3,C")  ' DEBUG
          TestBit3(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h5A ' BIT 3,D (Test bit 3 of register D)
          PrintDebug("BIT 3,D")  ' DEBUG
          TestBit3(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h5B ' BIT 3,E (Test bit 3 of register E)
          PrintDebug("BIT 3,E")  ' DEBUG
          TestBit3(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h5C ' BIT 3,H (Test bit 3 of register H)
          PrintDebug("BIT 3,H")  ' DEBUG
          TestBit3(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h5D ' BIT 3,L (Test bit 3 of register L)
          PrintDebug("BIT 3,L")  ' DEBUG
          TestBit3(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h5E ' BIT 3,(HL) (Test bit 3 of value at [HL])
          PrintDebug("BIT 3,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit3(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h5F ' BIT 3,A (Test bit 3 of register A)
          PrintDebug("BIT 3,A")  ' DEBUG
          TestBit3(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h60 ' BIT 4,B (Test bit 4 of register B)
          PrintDebug("BIT 4,B")  ' DEBUG
          TestBit4(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h61 ' BIT 4,C (Test bit 4 of register C)
          PrintDebug("BIT 4,C")  ' DEBUG
          TestBit4(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h62 ' BIT 4,D (Test bit 4 of register D)
          PrintDebug("BIT 4,D")  ' DEBUG
          TestBit4(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h63 ' BIT 4,E (Test bit 4 of register E)
          PrintDebug("BIT 4,E")  ' DEBUG
          TestBit4(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h64 ' BIT 4,H (Test bit 4 of register H)
          PrintDebug("BIT 4,H")  ' DEBUG
          TestBit4(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h65 ' BIT 4,L (Test bit 4 of register L)
          PrintDebug("BIT 4,L")  ' DEBUG
          TestBit4(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h66 ' BIT 4,(HL) (Test bit 4 of value at [HL])
          PrintDebug("BIT 4,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit4(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h67 ' BIT 4,A (Test bit 4 of register A)
          PrintDebug("BIT 4,A")  ' DEBUG
          TestBit4(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h68 ' BIT 5,B (Test bit 5 of register B)
          PrintDebug("BIT 5,B")  ' DEBUG
          TestBit5(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h69 ' BIT 5,C (Test bit 5 of register C)
          PrintDebug("BIT 5,C")  ' DEBUG
          TestBit5(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h6A ' BIT 5,D (Test bit 5 of register D)
          PrintDebug("BIT 5,D")  ' DEBUG
          TestBit5(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h6B ' BIT 5,E (Test bit 5 of register E)
          PrintDebug("BIT 5,E")  ' DEBUG
          TestBit5(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h6C ' BIT 5,H (Test bit 5 of register H)
          PrintDebug("BIT 5,H")  ' DEBUG
          TestBit5(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h6D ' BIT 5,L (Test bit 5 of register L)
          PrintDebug("BIT 5,L")  ' DEBUG
          TestBit5(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h6E ' BIT 5,(HL) (Test bit 5 of value at [HL])
          PrintDebug("BIT 5,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit5(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h6F ' BIT 5,A (Test bit 5 of register A)
          PrintDebug("BIT 5,A")  ' DEBUG
          TestBit5(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h70 ' BIT 6,B (Test bit 6 of register B)
          PrintDebug("BIT 6,B")  ' DEBUG
          TestBit6(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h71 ' BIT 6,C (Test bit 6 of register C)
          PrintDebug("BIT 6,C")  ' DEBUG
          TestBit6(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h72 ' BIT 6,D (Test bit 6 of register D)
          PrintDebug("BIT 6,D")  ' DEBUG
          TestBit6(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h73 ' BIT 6,E (Test bit 6 of register E)
          PrintDebug("BIT 6,E")  ' DEBUG
          TestBit6(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h74 ' BIT 6,H (Test bit 6 of register H)
          PrintDebug("BIT 6,H")  ' DEBUG
          TestBit6(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h75 ' BIT 6,L (Test bit 6 of register L)
          PrintDebug("BIT 6,L")  ' DEBUG
          TestBit6(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h76 ' BIT 6,(HL) (Test bit 6 of value at [HL])
          PrintDebug("BIT 6,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit6(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h77 ' BIT 6,A (Test bit 6 of register A)
          PrintDebug("BIT 6,A")  ' DEBUG
          TestBit6(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h78 ' BIT 7,B (Test bit 7 of register B)
          PrintDebug("BIT 7,B")  ' DEBUG
          TestBit7(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h79 ' BIT 7,C (Test bit 7 of register C)
          PrintDebug("BIT 7,C")  ' DEBUG
          TestBit7(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h7A ' BIT 7,D (Test bit 7 of register D)
          PrintDebug("BIT 7,D")  ' DEBUG
          TestBit7(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h7B ' BIT 7,E (Test bit 7 of register E)
          PrintDebug("BIT 7,E")  ' DEBUG
          TestBit7(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h7C ' BIT 7,H (Test bit 7 of register H)
          PrintDebug("BIT 7,H")  ' DEBUG
          TestBit7(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h7D ' BIT 7,L (Test bit 7 of register L)
          PrintDebug("BIT 7,L")  ' DEBUG
          TestBit7(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h7E ' BIT 7,(HL) (Test bit 7 of value at address HL)
          PrintDebug("BIT 7,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          TestBit7(UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h7F ' BIT 7,A (Test bit 7 of register A)
          PrintDebug("BIT 7,A")  ' DEBUG
          TestBit7(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h80 ' RES 0,B (Reset bit 0 of register B)
          PrintDebug("RES 0,B")  ' DEBUG
          ResetBit0(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h81 ' RES 0,C (Reset bit 0 of register C)
          PrintDebug("RES 0,C")  ' DEBUG
          ResetBit0(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h82 ' RES 0,D (Reset bit 0 of register D)
          PrintDebug("RES 0,D")  ' DEBUG
          ResetBit0(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h83 ' RES 0,E (Reset bit 0 of register E)
          PrintDebug("RES 0,E")  ' DEBUG
          ResetBit0(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h84 ' RES 0,H (Reset bit 0 of register H)
          PrintDebug("RES 0,H")  ' DEBUG
          ResetBit0(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h85 ' RES 0,L (Reset bit 0 of register L)
          PrintDebug("RES 0,L")  ' DEBUG
          ResetBit0(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h86 ' RES 0,(HL) (Reset bit 0 of value at [HL])
          PrintDebug("RES 0,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit0(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h87 ' RES 0,A (Reset bit 0 of register A)
          PrintDebug("RES 0,A")  ' DEBUG
          ResetBit0(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h88 ' RES 1,B (Reset bit 1 of register B)
          PrintDebug("RES 1,B")  ' DEBUG
          ResetBit1(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h89 ' RES 1,C (Reset bit 1 of register C)
          PrintDebug("RES 1,C")  ' DEBUG
          ResetBit1(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h8A ' RES 1,D (Reset bit 1 of register D)
          PrintDebug("RES 1,D")  ' DEBUG
          ResetBit1(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h8B ' RES 1,E (Reset bit 1 of register E)
          PrintDebug("RES 1,E")  ' DEBUG
          ResetBit1(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h8C ' RES 1,H (Reset bit 1 of register H)
          PrintDebug("RES 1,H")  ' DEBUG
          ResetBit1(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h8D ' RES 1,L (Reset bit 1 of register L)
          PrintDebug("RES 1,L")  ' DEBUG
          ResetBit1(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h8E ' RES 1,(HL) (Reset bit 1 of value at [HL])
          PrintDebug("RES 1,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit1(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h8F ' RES 1,A (Reset bit 1 of register A)
          PrintDebug("RES 1,A")  ' DEBUG
          ResetBit1(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h90 ' RES 2,B (Reset bit 2 of register B)
          PrintDebug("RES 2,B")  ' DEBUG
          ResetBit2(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h91 ' RES 2,C (Reset bit 2 of register C)
          PrintDebug("RES 2,C")  ' DEBUG
          ResetBit2(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h92 ' RES 2,D (Reset bit 2 of register D)
          PrintDebug("RES 2,D")  ' DEBUG
          ResetBit2(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h93 ' RES 2,E (Reset bit 2 of register E)
          PrintDebug("RES 2,E")  ' DEBUG
          ResetBit2(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h94 ' RES 2,H (Reset bit 2 of register H)
          PrintDebug("RES 2,H")  ' DEBUG
          ResetBit2(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h95 ' RES 2,L (Reset bit 2 of register L)
          PrintDebug("RES 2,L")  ' DEBUG
          ResetBit2(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h96 ' RES 2,(HL) (Reset bit 2 of value at [HL])
          PrintDebug("RES 2,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit2(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h97 ' RES 2,A (Reset bit 2 of register A)
          PrintDebug("RES 2,A")  ' DEBUG
          ResetBit2(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h98 ' RES 3,B (Reset bit 3 of register B)
          PrintDebug("RES 3,B")  ' DEBUG
          ResetBit3(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h99 ' RES 3,C (Reset bit 3 of register C)
          PrintDebug("RES 3,C")  ' DEBUG
          ResetBit3(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h9A ' RES 3,D (Reset bit 3 of register D)
          PrintDebug("RES 3,D")  ' DEBUG
          ResetBit3(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h9B ' RES 3,E (Reset bit 3 of register E)
          PrintDebug("RES 3,E")  ' DEBUG
          ResetBit3(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h9C ' RES 3,H (Reset bit 3 of register H)
          PrintDebug("RES 3,H")  ' DEBUG
          ResetBit3(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h9D ' RES 3,L (Reset bit 3 of register L)
          PrintDebug("RES 3,L")  ' DEBUG
          ResetBit3(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &h9E ' RES 3,(HL) (Reset bit 3 of value at [HL])
          PrintDebug("RES 3,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit3(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &h9F ' RES 3,A (Reset bit 3 of register A)
          PrintDebug("RES 3,A")  ' DEBUG
          ResetBit3(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA0 ' RES 4,B (Reset bit 4 of register B)
          PrintDebug("RES 4,B")  ' DEBUG
          ResetBit4(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA1 ' RES 4,C (Reset bit 4 of register C)
          PrintDebug("RES 4,C")  ' DEBUG
          ResetBit4(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA2 ' RES 4,D (Reset bit 4 of register D)
          PrintDebug("RES 4,D")  ' DEBUG
          ResetBit4(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA3 ' RES 4,E (Reset bit 4 of register E)
          PrintDebug("RES 4,E")  ' DEBUG
          ResetBit4(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA4 ' RES 4,H (Reset bit 4 of register H)
          PrintDebug("RES 4,H")  ' DEBUG
          ResetBit4(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA5 ' RES 4,L (Reset bit 4 of register L)
          PrintDebug("RES 4,L")  ' DEBUG
          ResetBit4(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA6 ' RES 4,(HL) (Reset bit 4 of value at [HL])
          PrintDebug("RES 4,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit4(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hA7 ' RES 4,A (Reset bit 4 of register A)
          PrintDebug("RES 4,A")  ' DEBUG
          ResetBit4(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA8 ' RES 5,B (Reset bit 5 of register B)
          PrintDebug("RES 5,B")  ' DEBUG
          ResetBit5(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hA9 ' RES 5,C (Reset bit 5 of register C)
          PrintDebug("RES 5,C")  ' DEBUG
          ResetBit5(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hAA ' RES 5,D (Reset bit 5 of register D)
          PrintDebug("RES 5,D")  ' DEBUG
          ResetBit5(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hAB ' RES 5,E (Reset bit 5 of register E)
          PrintDebug("RES 5,E")  ' DEBUG
          ResetBit5(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hAC ' RES 5,H (Reset bit 5 of register H)
          PrintDebug("RES 5,H")  ' DEBUG
          ResetBit5(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hAD ' RES 5,L (Reset bit 5 of register L)
          PrintDebug("RES 5,L")  ' DEBUG
          ResetBit5(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hAE ' RES 5,(HL) (Reset bit 5 of value at [HL])
          PrintDebug("RES 5,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit5(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hAF ' RES 5,A (Reset bit 5 of register A)
          PrintDebug("RES 5,A")  ' DEBUG
          ResetBit5(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB0 ' RES 6,B (Reset bit 6 of register B)
          PrintDebug("RES 6,B")  ' DEBUG
          ResetBit6(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB1 ' RES 6,C (Reset bit 6 of register C)
          PrintDebug("RES 6,C")  ' DEBUG
          ResetBit6(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB2 ' RES 6,D (Reset bit 6 of register D)
          PrintDebug("RES 6,D")  ' DEBUG
          ResetBit6(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB3 ' RES 6,E (Reset bit 6 of register E)
          PrintDebug("RES 6,E")  ' DEBUG
          ResetBit6(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB4 ' RES 6,H (Reset bit 6 of register H)
          PrintDebug("RES 6,H")  ' DEBUG
          ResetBit6(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB5 ' RES 6,L (Reset bit 6 of register L)
          PrintDebug("RES 6,L")  ' DEBUG
          ResetBit6(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB6 ' RES 6,(HL) (Reset bit 6 of value at [HL])
          PrintDebug("RES 6,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit6(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hB7 ' RES 6,A (Reset bit 6 of register A)
          PrintDebug("RES 6,A")  ' DEBUG
          ResetBit6(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB8 ' RES 7,B (Reset bit 7 of register B)
          PrintDebug("RES 7,B")  ' DEBUG
          ResetBit7(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hB9 ' RES 7,C (Reset bit 7 of register C)
          PrintDebug("RES 7,C")  ' DEBUG
          ResetBit7(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hBA ' RES 7,D (Reset bit 7 of register D)
          PrintDebug("RES 7,D")  ' DEBUG
          ResetBit7(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hBB ' RES 7,E (Reset bit 7 of register E)
          PrintDebug("RES 7,E")  ' DEBUG
          ResetBit7(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hBC ' RES 7,H (Reset bit 7 of register H)
          PrintDebug("RES 7,H")  ' DEBUG
          ResetBit7(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hBD ' RES 7,L (Reset bit 7 of register L)
          PrintDebug("RES 7,L")  ' DEBUG
          ResetBit7(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hBE ' RES 7,(HL) (Reset bit 7 of value at [HL])
          PrintDebug("RES 7,(HL)")  ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          ResetBit7(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hBF ' RES 7,A (Reset bit 7 of register A)
          PrintDebug("RES 7,A")  ' DEBUG
          ResetBit7(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC0  ' SET 0,B  (Set bit 0 of B)
          PrintDebug("SET 0,B")   ' DEBUG
          SetBit0(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC1  ' SET 0,C  (Set bit 0 of C)
          PrintDebug("SET 0,C")   ' DEBUG
          SetBit0(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC2  ' SET 0,D  (Set bit 0 of D)
          PrintDebug("SET 0,D")   ' DEBUG
          SetBit0(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC3  ' SET 0,E  (Set bit 0 of E)
          PrintDebug("SET 0,E")   ' DEBUG
          SetBit0(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC4  ' SET 0,H  (Set bit 0 of H)
          PrintDebug("SET 0,H")   ' DEBUG
          SetBit0(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC5  ' SET 0,L  (Set bit 0 of L)
          PrintDebug("SET 0,L")   ' DEBUG
          SetBit0(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC6  ' SET 0,(HL)  (Set bit 0 of value at HL)
          PrintDebug("SET 0,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit0(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hC7  ' SET 0,A  (Set bit 0 of A)
          PrintDebug("SET 0,A")   ' DEBUG
          SetBit0(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC8  ' SET 1,B  (Set bit 1 of B)
          PrintDebug("SET 1,B")   ' DEBUG
          SetBit1(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hC9  ' SET 1,C  (Set bit 1 of C)
          PrintDebug("SET 1,C")   ' DEBUG
          SetBit1(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hCA  ' SET 1,D  (Set bit 1 of D)
          PrintDebug("SET 1,D")   ' DEBUG
          SetBit1(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hCB  ' SET 1,E  (Set bit 1 of E)
          PrintDebug("SET 1,E")   ' DEBUG
          SetBit1(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hCC  ' SET 1,H  (Set bit 1 of H)
          PrintDebug("SET 1,H")   ' DEBUG
          SetBit1(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hCD  ' SET 1,L  (Set bit 1 of L)
          PrintDebug("SET 1,L")   ' DEBUG
          SetBit1(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hCE  ' SET 1,(HL)  (Set bit 1 of value at HL)
          PrintDebug("SET 1,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit1(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hCF  ' SET 1,A  (Set bit 1 of A)
          PrintDebug("SET 1,A")   ' DEBUG
          SetBit1(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD0  ' SET 2,B  (Set bit 2 of B)
          PrintDebug("SET 2,B")   ' DEBUG
          SetBit2(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD1  ' SET 2,C  (Set bit 2 of C)
          PrintDebug("SET 2,C")   ' DEBUG
          SetBit2(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD2  ' SET 2,D  (Set bit 2 of D)
          PrintDebug("SET 2,D")   ' DEBUG
          SetBit2(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD3  ' SET 2,E  (Set bit 2 of E)
          PrintDebug("SET 2,E")   ' DEBUG
          SetBit2(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD4  ' SET 2,H  (Set bit 2 of H)
          PrintDebug("SET 2,H")   ' DEBUG
          SetBit2(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD5  ' SET 2,L  (Set bit 2 of L)
          PrintDebug("SET 2,L")   ' DEBUG
          SetBit2(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD6  ' SET 2,(HL)  (Set bit 2 of value at HL)
          PrintDebug("SET 2,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit2(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hD7  ' SET 2,A  (Set bit 2 of A)
          PrintDebug("SET 2,A")   ' DEBUG
          SetBit2(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD8  ' SET 3,B  (Set bit 3 of B)
          PrintDebug("SET 3,B")   ' DEBUG
          SetBit3(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hD9  ' SET 3,C  (Set bit 3 of C)
          PrintDebug("SET 3,C")   ' DEBUG
          SetBit3(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hDA  ' SET 3,D  (Set bit 3 of D)
          PrintDebug("SET 3,D")   ' DEBUG
          SetBit3(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hDB  ' SET 3,E  (Set bit 3 of E)
          PrintDebug("SET 3,E")   ' DEBUG
          SetBit3(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hDC  ' SET 3,H  (Set bit 3 of H)
          PrintDebug("SET 3,H")   ' DEBUG
          SetBit3(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hDD  ' SET 3,L  (Set bit 3 of L)
          PrintDebug("SET 3,L")   ' DEBUG
          SetBit3(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hDE  ' SET 3,(HL)  (Set bit 3 of value at HL)
          PrintDebug("SET 3,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit3(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hDF  ' SET 3,A  (Set bit 3 of A)
          PrintDebug("SET 3,A")   ' DEBUG
          SetBit3(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE0  ' SET 4,B  (Set bit 4 of B)
          PrintDebug("SET 4,B")   ' DEBUG
          SetBit4(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE1  ' SET 4,C  (Set bit 4 of C)
          PrintDebug("SET 4,C")   ' DEBUG
          SetBit4(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE2  ' SET 4,D  (Set bit 4 of D)
          PrintDebug("SET 4,D")   ' DEBUG
          SetBit4(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE3  ' SET 4,E  (Set bit 4 of E)
          PrintDebug("SET 4,E")   ' DEBUG
          SetBit4(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE4  ' SET 4,H  (Set bit 4 of H)
          PrintDebug("SET 4,H")   ' DEBUG
          SetBit4(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE5  ' SET 4,L  (Set bit 4 of L)
          PrintDebug("SET 4,L")   ' DEBUG
          SetBit4(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE6  ' SET 4,(HL)  (Set bit 4 of value at HL)
          PrintDebug("SET 4,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit4(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hE7  ' SET 4,A  (Set bit 4 of A)
          PrintDebug("SET 4,A")   ' DEBUG
          SetBit4(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE8  ' SET 5,B  (Set bit 5 of B)
          PrintDebug("SET 5,B")   ' DEBUG
          SetBit5(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hE9  ' SET 5,C  (Set bit 5 of C)
          PrintDebug("SET 5,C")   ' DEBUG
          SetBit5(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hEA  ' SET 5,D  (Set bit 5 of D)
          PrintDebug("SET 5,D")   ' DEBUG
          SetBit5(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hEB  ' SET 5,E  (Set bit 5 of E)
          PrintDebug("SET 5,E")   ' DEBUG
          SetBit5(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hEC  ' SET 5,H  (Set bit 5 of H)
          PrintDebug("SET 5,H")   ' DEBUG
          SetBit5(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hED  ' SET 5,L  (Set bit 5 of L)
          PrintDebug("SET 5,L")   ' DEBUG
          SetBit5(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hEE  ' SET 5,(HL)  (Set bit 5 of value at HL)
          PrintDebug("SET 5,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit5(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hEF  ' SET 5,A  (Set bit 5 of A)
          PrintDebug("SET 5,A")   ' DEBUG
          SetBit5(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF0  ' SET 6,B  (Set bit 6 of B)
          PrintDebug("SET 6,B")   ' DEBUG
          SetBit6(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF1  ' SET 6,C  (Set bit 6 of C)
          PrintDebug("SET 6,C")   ' DEBUG
          SetBit6(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF2  ' SET 6,D  (Set bit 6 of D)
          PrintDebug("SET 6,D")   ' DEBUG
          SetBit6(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF3  ' SET 6,E  (Set bit 6 of E)
          PrintDebug("SET 6,E")   ' DEBUG
          SetBit6(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF4  ' SET 6,H  (Set bit 6 of H)
          PrintDebug("SET 6,H")   ' DEBUG
          SetBit6(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF5  ' SET 6,L  (Set bit 6 of L)
          PrintDebug("SET 6,L")   ' DEBUG
          SetBit6(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF6  ' SET 6,(HL)  (Set bit 6 of value at HL)
          PrintDebug("SET 6,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit6(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hF7  ' SET 6,A  (Set bit 6 of A)
          PrintDebug("SET 6,A")   ' DEBUG
          SetBit6(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF8  ' SET 7,B  (Set bit 7 of B)
          PrintDebug("SET 7,B")   ' DEBUG
          SetBit7(Register.B)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hF9  ' SET 7,C  (Set bit 7 of C)
          PrintDebug("SET 7,C")   ' DEBUG
          SetBit7(Register.C)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hFA  ' SET 7,D  (Set bit 7 of D)
          PrintDebug("SET 7,D")   ' DEBUG
          SetBit7(Register.D)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hFB  ' SET 7,E  (Set bit 7 of E)
          PrintDebug("SET 7,E")   ' DEBUG
          SetBit7(Register.E)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hFC  ' SET 7,H  (Set bit 7 of H)
          PrintDebug("SET 7,H")   ' DEBUG
          SetBit7(Register.H)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hFD  ' SET 7,L  (Set bit 7 of L)
          PrintDebug("SET 7,L")   ' DEBUG
          SetBit7(Register.L)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE &hFE  ' SET 7,(HL)  (Set bit 7 of value at HL)
          PrintDebug("SET 7,(HL)")   ' DEBUG
          MemoryRead(ReadRegHL(), UbyteBuff1)
          SetBit7(UbyteBuff1)
          MemoryWrite(ReadRegHL(), UbyteBuff1)
          Register.PC += 2
          UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
        CASE &hFF  ' SET 7,A  (Set bit 7 of A)
          PrintDebug("SET 7,A")   ' DEBUG
          SetBit7(Register.A)
          Register.PC += 2
          UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
        CASE ELSE
          PrintDebug("Unknown opcode at " & HEX(Register.PC) & ": " & HEX(CpuInstruction, 4))
          QuitEmulator = 1
      END SELECT
    CASE &hCC   ' CALL Z,nn (jump if Z=1)
      PrintDebug("CALL Z,nn") ' DEBUG
      IF GetFlagZ() = 0 THEN   ' Don't jump
          Register.PC += 3
        ELSE  ' Jump
          PushToStack16(Register.PC + 3)   ' First save the current value of the program counter (+3 bytes)
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hCD   ' CALL (calls a subroutine)
      PushToStack16(Register.PC + 3)    ' First save the current value of the program counter (+3 bytes)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("CALL " & HEX(DwordVal(UbyteBuff1, UbyteBuff2))) ' DEBUG
      Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hCE   ' ADC A,n
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("ADC A," & HEX(UbyteBuff1,2))   ' DEBUG
      AdcA(UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hCF   ' RST 08h
      PrintDebug("RST 08h")   ' DEBUG
      PushToStack16(Register.PC + 1)           ' Save current address (+1)
      Register.PC = &h08  ' Jump to 08h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hD0   ' RET NC   ' RET if C = 0
      PrintDebug("RET NC")
      IF GetFlagC() = 0 THEN
          PopPCfromStack()
        ELSE
          Register.PC += 1
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hD1   ' POP DE (Write the value in stack into DE)
      PrintDebug("POP DE")  ' DEBUG
      PopFromStack(Register.D, Register.E)
      Register.PC += 1
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hD2   ' JP NC,nn (jump if C=0)
      PrintDebug("JP NC,nn") ' DEBUG
      IF GetFlagC() = 0 THEN  ' Jump
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hD4   ' CALL NC,nn (jump if Z=0)
      PrintDebug("CALL NC,nn") ' DEBUG
      IF GetFlagC() = 0 THEN  ' Jump
          PushToStack16(Register.PC + 3)    ' First save the current value of the program counter (+3 bytes)
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hD5   ' PUSH DE
      PrintDebug("PUSH DE")  ' DEBUG
      PushToStack(Register.D, Register.E)
      Register.PC += 1
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hD6  ' SUB n from A
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("SUB " & HEX(UbyteBuff1))  ' DEBUG
      SubValFromReg8(Register.A, UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hD7   ' RST 10h
      PrintDebug("RST 10h")   ' DEBUG
      PushToStack16(Register.PC + 1)     ' Save current address (+1)
      Register.PC = &h10  ' Jump to 10h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hD8   ' RET C   ' RET if C = 1
      PrintDebug("RET C")
      IF GetFlagC() = 0 THEN
          Register.PC += 1 
        ELSE
          PopPCfromStack()
      END IF
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hD9   ' RETI  (RET + EI)
      PrintDebug("RETI")   ' DEBUG
      PopPCfromStack()           ' RET (return from a call or int)
      InterruptsState = 1        ' EI (enable interrupts)
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hDA   ' JP C,nn (jump if C=1)
      PrintDebug("JP C,nn") ' DEBUG
      IF GetFlagC() = 0 THEN  ' Don't jump
          Register.PC += 3
        ELSE   ' Jump
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hDC   ' CALL C,nn (jump if Z=1)
      PrintDebug("CALL C,nn") ' DEBUG
      IF GetFlagC() = 1 THEN  ' Jump
          PushToStack16(Register.PC + 3)    ' First save the current value of the program counter (+3 bytes)
          MemoryRead(Register.PC + 1, UbyteBuff1)
          MemoryRead(Register.PC + 2, UbyteBuff2)
          Register.PC = DwordVal(UbyteBuff1, UbyteBuff2)
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hDE  ' SBC A,n
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("SBC A," & UbyteBuff1)  ' DEBUG
      SbcA(UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hDF   ' RST 18h
      PrintDebug("RST 18h")   ' DEBUG
      PushToStack16(Register.PC + 1)     ' Save current address (+1)
      Register.PC = &h18  ' Jump to 18h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hE0   ' LD ($FF00+n),A
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("LD ($FF00+" & HEX(UbyteBuff1, 2) & "),A")  ' DEBUG
      MemoryWrite(&hFF00 + UbyteBuff1, Register.A)
      Register.PC += 2
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hE1   ' POP HL (Write the value in stack into HL)
      PrintDebug("POP HL")  ' DEBUG
      PopFromStack(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hE2   ' LD (FF00h+C),A (put A into addr $FF00+C)
      PrintDebug("LD (" & HEX(&hFF00 + Register.C) & "),A")   ' DEBUG
      MemoryWrite(&hFF00 + Register.C, Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hE5   ' PUSH HL
      PrintDebug("PUSH HL")  ' DEBUG
      PushToStack(Register.H, Register.L)
      Register.PC += 1
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hE6   ' A AND n
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("A AND " & UbyteBuff1)
      AndA(UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hE7   ' RST 20h
      PrintDebug("RST 20h")   ' DEBUG
      PushToStack16(Register.PC + 1)    ' Save current address (+1)
      Register.PC = &h20  ' Jump to 20h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hE8   ' ADD SP,n
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("ADD SP," & UbyteToByte(UbyteBuff1)) ' DEBUG
      AddToSP(UbyteToByte(UbyteBuff1))
      Register.PC += 2
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hE9   ' JP (HL) / Jumps to address contained in HL (same as "LD PC,HL")
      PrintDebug("JP (HL)")  ' DEBUG
      Register.PC = ReadRegHL()
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hEA   ' LD (nn),A   ' Load value of A into address nn
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD (" & HEX(DwordVal(UbyteBuff1, UbyteBuff2)) & "),A")  ' DEBUG
      MemoryWrite(DwordVal(UbyteBuff1, UbyteBuff2), Register.A)
      Register.PC += 3
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hEE   ' XOR n  (A = A XOR n)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("XOR " & HEX(UbyteBuff1, 2)) ' DEBUG
      XorA(UbyteBuff1) ' Xor with A, result in A
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hEF   ' RST 28h
      PrintDebug("RST 28h")   ' DEBUG
      PushToStack16(Register.PC + 1)    ' Save current address (+1)
      Register.PC = &h28  ' Jump to 28h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hF0   ' LD A,($FF00+n)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("LD A,($FF00+" & HEX(UbyteBuff1, 2) & ")")  ' DEBUG
      MemoryRead(&hFF00 + UbyteBuff1, Register.A)
      Register.PC += 2
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hF1   ' POP AF (Write the value in stack into AF)
      PrintDebug("POP AF")  ' DEBUG
      PopFromStack(Register.A, Register.F)
      Register.PC += 1
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hF2   ' LD A,(FF00h+C)  (put value at address FF00+C into A)
      PrintDebug("LD A,(FF00+C)")    ' DEBUG
      MemoryRead(&hFF00 + Register.C, Register.A)
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hF3   ' DI (Disables Interrupts)
      PrintDebug("DI - Disables interrupts")  ' DEBUG
      InterruptsState = 0   ' Sets IME (Interrupts Master Enable) to 0
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hF5   ' PUSH AF
      PrintDebug("PUSH AF")  ' DEBUG
      PushToStack(Register.A, Register.F)
      Register.PC += 1
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hF6   ' OR n (A = A OR n)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("OR " & HEX(UbyteBuff1, 2))  ' DEBUG
      OrA(UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hF7   ' RST 30h
      PrintDebug("RST 30h")   ' DEBUG
      PushToStack16(Register.PC + 1)     ' Save current address (+1)
      Register.PC = &h30  ' Jump to 30h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE &hF8   ' LD HL,SP+n
      MemoryRead(Register.PC + 1, UbyteBuff1)
      ByteBuff = UbyteToByte(UbyteBuff1)
      PrintDebug("LD HL,SP+" & ByteBuff)
      UintBuff = ((Register.SP + ByteBuff) AND &hFFFF)
      IF (Register.SP > UintBuff) THEN SetFlagC() ELSE ResetFlagC()
      IF (((Register.SP XOR ByteBuff XOR UintBuff) AND 4096) > 0) THEN
          SetFlagH()
        ELSE
          ResetFlagH()
      END IF
      WriteRegHL((UintBuff SHR 8), (UintBuff AND &b11111111))
      ResetFlagZ()
      ResetFlagN()
      Register.PC += 2
      UsedCycles = 12  ' that many CPU cycles should be spent on this instruction
    CASE &hF9   ' LD SP,HL  (put HL into SP)
      PrintDebug("LD SP,HL")
      Register.SP = ReadRegHL()
      Register.PC += 1
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hFA   ' LD A,nn  (put value from address nn to A)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      MemoryRead(Register.PC + 2, UbyteBuff2)
      PrintDebug("LD A," & HEX(DwordVal(UbyteBuff1, UbyteBuff2), 4))
      MemoryRead(DwordVal(UbyteBuff1, UbyteBuff2), Register.A)
      Register.PC += 3
      UsedCycles = 16  ' that many CPU cycles should be spent on this instruction
    CASE &hFB   ' EI (Enables Interrupts)
      PrintDebug("EI - Enables interrupts")  ' DEBUG
      InterruptsState = 1
      Register.PC += 1
      UsedCycles = 4  ' that many CPU cycles should be spent on this instruction
    CASE &hFE   ' CP n (compare A with n)
      MemoryRead(Register.PC + 1, UbyteBuff1)
      PrintDebug("CP A," & HEX(UbyteBuff1, 2)) ' DEBUG
      CmpA(UbyteBuff1)
      Register.PC += 2
      UsedCycles = 8  ' that many CPU cycles should be spent on this instruction
    CASE &hFF   ' RST 38h
      PrintDebug("RST 38h")   ' DEBUG
      PushToStack16(Register.PC + 1)     ' Save current address (+1)
      Register.PC = &h38  ' Jump to 38h
      UsedCycles = 32  ' that many CPU cycles should be spent on this instruction
    CASE ELSE
      PrintMsg("UNKNOWN OPCODE AT " & HEX(Register.PC, 4) & ": " & HEX(CpuInstruction, 2))
      #IFDEF DEBUGMODE
        PrintDebug("Got an unknown opcode! -> " & HEX(CpuInstruction, 2))
        DebugOnScreen()
      #ENDIF
      PressAnyKey()
      QuitEmulator = 1
  END SELECT
  #IFDEF DEBUGMODE
    DIM AS STRING LastKey
    IF (DebugCpuPause = 1) THEN
        DebugOnscreen()
        DebugShowCpuRegisters()
        DO                                                      ' DEBUG
          SLEEP 10,1                                            ' DEBUG
          LastKey = INKEY
        LOOP UNTIL (LEN(LastKey) > 0)                           ' DEBUG
      ELSE
        LastKey = INKEY
    END IF
    IF LastKey = " " THEN DebugCpuPause = 1
    IF LastKey = "p" THEN
      DebugOnscreen()
      DebugShowCpuRegisters()
      DO                                                      ' DEBUG
        SLEEP 10,1                                            ' DEBUG
        LastKey = INKEY
      LOOP UNTIL (LEN(LastKey) > 0)                           ' DEBUG
    END IF
  #ENDIF
  'IF UsedCycles = 0 THEN PrintMsg("ERR: CYCLES VALUE IS 0!" & CHR(10) & "LAST INSTRUCTION: " & HEX(CpuInstruction) & "H") : SLEEP : END
  TotalCycles += UsedCycles     ' Increment the global cycles counter
  uTimer(UsedCycles)            ' Update uTimer
  incDivider(UsedCycles)        ' Increment the DIV register
  VideoSysUpdate(UsedCycles)    ' Update Video subsystem
END SUB
