REM
REM --------------------------------------
REM This file is part of the zBoy project.
REM Copyright (C) Mateusz Viste 2010
REM --------------------------------------
REM
REM * MBC5 short documentation *
REM
REM The MBC5 controller is similar to the MBC3 (but no RTC) but can access up to 64 Mbits of ROM and up to
REM 1 Mbit of RAM. The lower 8 bits of the 9-bit ROM bank select is written to the 2000-2FFF area while the
REM upper bit is written to the least significant bit of the 3000-3FFF area.
REM
REM Writing a value (XXXXBBBB - X = Don't care, B = bank select bits) into 4000-5FFF area will select an
REM appropriate RAM bank at A000-BFFF if the cart contains RAM. Ram sizes are 64 Kbits, 256 Kbits, and 1 Mbit.
REM
REM Also, this is the first MBC that allows ROM bank 0 to appear in the 4000-7FFF range by writing $000 to the
REM rom bank select.
REM

SUB MBC5_MemoryRead(BYREF memaddr AS UINTEGER, BYREF Result AS UBYTE)
  IF (memaddr <= &h00FF) AND (BootRomEnabledFlag = 1) THEN
    Result = BootROM(memaddr)
  ELSEIF (memaddr >= &hA000) AND (memaddr < &hC000) THEN        ' RAM bank n
    Result = MemoryBankedRAM((CurRamBank SHL 13) + memaddr)
  ELSEIF (memaddr >= &hC000) AND (memaddr < &hE000) THEN    ' Internal 8KiB RAM
    Result = MemoryInternalRAM(memaddr)
  ELSEIF (memaddr >= &hE000) AND (memaddr < &hFE00) THEN    ' RAM mirror
    Result = MemoryInternalRAM(memaddr - 8192)
  ELSEIF (memaddr >= &hFE00) AND (memaddr < &hFEA0) THEN    ' Sprite OAM memory
    Result = SpriteOAM(memaddr)
  ELSEIF (memaddr >= &hFF80) AND (memaddr <= &hFFFF) THEN   ' Hi RAM area
    Result = MemoryInternalHiRAM(memaddr)
  ELSEIF (memaddr >= &h0000) AND (memaddr < &h4000) THEN    ' ROM bank #0
    Result = MemoryROM(memaddr)
  ELSEIF (memaddr >= &hFF00) AND (memaddr <= &hFF4B) THEN   ' I/O registers
    Result = IoRegisters(memaddr)
  ELSEIF (memaddr >= &h4000) AND (memaddr < &h8000) THEN    ' ROM bank #n
    IF ((CurRomBank SHL 14) + (memaddr - &h4000)) >= RomInfos.RomSize THEN  ' SHL 14 is the same than *16384 (but faster)
      PrintMsg("INVALID ROM ADDRESS!" & CHR(10) & "CURBANK=" & CurRomBank)
      QuitEmulator = 1
      PressAnyKey()
    END IF
    Result = MemoryROM((CurRomBank SHL 14) + (memaddr - &h4000))
  ELSEIF (memaddr >= &h8000) AND (memaddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    Result = VideoRAM(memaddr)
  ELSE                                                  ' Else it's something I don't know about
    'PRINT "MMU fatal error: Tried to read from an unknown memory address: " & HEX(memaddr) & "h"
    'SLEEP
    'QuitEmulator = 1
    Result = MemoryMAP(memaddr)
  END IF
END SUB


SUB MBC5_MemoryWrite(BYREF memaddr AS UINTEGER, BYREF DataByte AS UBYTE)
  IF (memaddr < &h1FFF) THEN  ' Enables / disables current RAM bank (not supported yet)
    REM Before using any RAM bank, you need to enable it by writing xxxx1010 into 0000-1FFF (writing there
    REM any non-xxxx1010 value disables the RAM bank).
  ELSEIF (memaddr >= &h2000) AND (memaddr < &h3000) THEN     ' Select current ROM bank on MBC5 (8 lowest bits)
    CurRomBank = (CurRomBank AND &b100000000) OR DataByte
  ELSEIF (memaddr >= &h3000) AND (memaddr < &h4000) THEN     ' Select current ROM bank on MBC5 (most significant bit)
    IF (DataByte AND &b00000001) = 0 THEN
        CurRomBank = CurRomBank AND &b011111111   ' Reset the most significant bit of the 9bit bank select
      ELSE
        CurRomBank = CurRomBank OR &b100000000    ' Set the most significant bit of the 9bit bank select
    END IF
  ELSEIF (memaddr >= &h4000) AND (memaddr < &h6000) THEN  ' Select RAM bank (xxxxBBBB)
    CurRamBank = (DataByte AND &b00001111)
  ELSEIF (memaddr >= &hA000) AND (memaddr < &hC000) THEN    ' RAM bank n
    MemoryBankedRAM((CurRamBank SHL 13) + memaddr) = DataByte
  ELSEIF (memaddr >= &hC000) AND (memaddr < &hE000) THEN    ' Internal 8KiB RAM
    MemoryInternalRAM(memaddr) = DataByte
  ELSEIF (memaddr >= &hE000) AND (memaddr < &hFE00) THEN    ' RAM mirror
    MemoryInternalRAM(memaddr - 8192) = DataByte
  ELSEIF (memaddr >= &hFE00) AND (memaddr < &hFEA0) THEN    ' Sprite OAM memory
    'IF (SpriteOAM(memaddr) <> DataByte) THEN  ' If value is the same, do nothing
      SpriteOAM(memaddr) = DataByte
    'END IF
  ELSEIF (memaddr >= &hFF80) AND (memaddr <= &hFFFF) THEN   ' Hi RAM area
    MemoryInternalHiRAM(memaddr) = DataByte
  ELSEIF (memaddr = &hFF41) THEN ' STAT register: Do not allow to write into 2 last bits of the STAT register, as these
    IoRegisters(&hFF41) = (IoRegisters(&hFF41) AND &b00000011) OR (DataByte AND &b11111100)    ' bits are the mode flag.
  ELSEIF (memaddr = &hFF44) THEN ' CURLINE [RW] Current Scanline.
    IoRegisters(memaddr) = 0     ' Writing into this register resets it.
  ELSEIF (memaddr = &hFF46) THEN ' Starts LCD OAM DMA transfer
    DIM AS INTEGER x             '
    IF DataByte <= &hF1 THEN     ' Normally, should take 160us... but who cares?
        FOR x = 0 TO 159         ' Let's copy XX00-XX9F to FE00-FE9F
          MBC5_MemoryRead((DataByte SHL 8) + x, SpriteOAM(&hFE00 + x)) ' SHL 8 is the same than * 256 (but faster)
        NEXT x
      ELSE
        PrintMsg("MMU FATAL ERROR! TRIED TO RUN LCD OAM DMA TRANSFER USING AN INVALID RANGE!")
        PressAnyKey()
        END
    END IF
  ELSEIF(memaddr = &hFF04) THEN
    IoRegisters(&hFF04) = 0      ' Divide register: Writing any value sets it to 0
  ELSEIF (memaddr >= &hFF00) AND (memaddr <= &hFF4B) THEN   ' I/O registers
    'IF (IoRegisters(memaddr) <> DataByte) THEN ' Do nothing if value don't change
      IoRegisters(memaddr) = DataByte
    'END IF
  ELSEIF (memaddr = &hFF50) THEN
    BootRomEnabledFlag = 0
  ELSEIF (memaddr >= &h8000) AND (memaddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    'IF (VideoRAM(memaddr) <> DataByte) THEN  ' If value is not different, then do nothing
      VideoRAM(memaddr) = DataByte
    'END IF
  'ELSEIF (memaddr >= &hFEA0) AND (memaddr <= &hFEFF) THEN   ' This area is not usable, but some games writes to it...
  '  REM                                                     ' For example Dr. Mario tries to write to FEFF... who knows why?
  'ELSEIF (memaddr >= &hFF4C) AND (memaddr <= &hFF7F) THEN   ' This area is not usable, but some games writes to it...
  '  REM                                                     ' For example Dr. Mario tries to write to FF7F... who knows why?
  ELSE                                                      ' Else it's something I don't know about
    'PRINT "MMU fatal error: Tried to write to an unknown memory address: "; HEX(memaddr)
    'SLEEP
    'QuitEmulator = 1
    MemoryMAP(memaddr) = DataByte
  END IF
END SUB
